@@ -304,19 +304,28 @@ opt_milter_version (struct mu_parseopt *po, struct mu_option *opt,
   char *p;
   unsigned long maj, min, pat;
 
+  errno = 0;
   maj = strtoul (arg, &p, 0);
-  if (*p == 0)
+  if (errno != 0 || p == arg)
+    /* error */;
+  else if (*p == 0)
     {
       milter_version_option = maj;
       return;
     }
   else if (*p == '.')
     {
-      min = strtoul (p + 1, &p, 0);
-      if (*p == '.')
+      arg = p + 1;
+      min = strtoul (arg, &p, 0);
+      if (errno != 0 || p == arg)
+	/* error */;
+      else if (*p == '.')
 	{
-	  pat = strtoul (p + 1, &p, 0);
-	  if (*p == 0)
+	  arg = p + 1;
+	  pat = strtoul (arg, &p, 0);
+	  if (errno != 0 || p == arg)
+	    /* error */;
+	  else if (*p == 0)
 	    {
 	      milter_version_option = GACOPYZ_SM_MKVER (maj, min, pat);
 	      return;
@@ -332,6 +341,38 @@ opt_milter_version (struct mu_parseopt *po, struct mu_option *opt,
   exit (po->po_exit_error);
 }
 
+static int
+strtotv (char const *arg, char **endp, struct timeval *rettv)
+{
+  char *p;
+  struct timeval tv;
+
+  errno = 0;
+  tv.tv_sec = strtoul (arg, &p, 10);
+  if (errno)
+    goto err;
+  else if (p == arg)
+    {
+      errno = EINVAL;
+      goto err;
+    }
+  if (*p == '.')
+    {
+      errno = 0;
+      tv.tv_usec = strtoul (p + 1, &p, 10);
+      if (errno)
+	goto err;
+    }
+  else
+    tv.tv_usec = 0;
+
+  *rettv = tv;
+ err:
+  if (endp)
+    *endp = p;
+  return errno ? -1 : 0;
+}
+
 static void
 opt_milter_timeout (struct mu_parseopt *po, struct mu_option *opt,
 		    char const *arg)
@@ -377,22 +418,31 @@ opt_milter_timeout (struct mu_parseopt *po, struct mu_option *opt,
 	  mu_parseopt_error (po, _("missing '=' after %c"), *arg);
 	  exit (po->po_exit_error);
 	}
-      tv.tv_sec = strtoul (arg + 1, &p, 10);
-      if (*p == '.')
-	tv.tv_usec = strtoul (p + 1, &p, 10);
-      while (*p && isspace (*p))
-	p++;
-      if (*p == ',')
-	arg = p + 1;
-      else if (*p == 0)
-	arg = p;
+
+      arg++;
+      if (strtotv (arg, &p, &tv) == 0)
+	{
+	  milter_timeouts[ind] = tv;
+	  while (*p && isspace (*p))
+	    p++;
+	  if (*p == ',')
+	    arg = p + 1;
+	  else if (*p == 0)
+	    break;
+	  else
+	    {
+	      mu_parseopt_error (po, _("unexpected character at %s"), p);
+	      exit (po->po_exit_error);
+	    }
+	}
       else
 	{
-	  mu_parseopt_error (po, _("unexpected character: %c"), *p);
+	  if (errno == EINVAL)
+	    mu_parseopt_error (po, _("invalid timeout near %s"), p);
+	  else
+	    mu_parseopt_error (po, _("value out of range: %s"), arg);
 	  exit (po->po_exit_error);
 	}
-      tv.tv_usec = 0;
-      milter_timeouts[ind] = tv;
     }
 }
 
@@ -2472,6 +2522,8 @@ shell_help ()
     " \\L[name] [name...]                 List macros",
     " \\Uname [name...]                   Undefine Sendmail macros",
     " \\Sfamily hostname address [port]   Define sender socket address",
+    " \\Wseconds[.fraq]                   Wait given number of seconds",
+    " \\#text                             Comment",
   };
   int hcount = sizeof (hstr) / sizeof (hstr[0]);
   int i;
@@ -2487,6 +2539,9 @@ shell (int argc, char **argv)
 {
   switch (argv[0][1])
     {
+    case '#':
+      break;
+
     case 'd':
     case 'D':
       {
@@ -2542,6 +2597,20 @@ shell (int argc, char **argv)
       /* \Sfamily hostname address [port] */
       set_sender_socket (argc, argv);
       break;
+
+    case 'w':
+    case 'W':
+      /* \Wseconds */
+      {
+	struct timeval tv;
+	if (strtotv (argv[0] + 2, NULL, &tv) == 0)
+	  {
+	    select (0, NULL, NULL, NULL, &tv);
+	  }
+	else
+	  smtp_reply (502, "Bad time to wait");
+      }
+      break;
       
     case '?':
       shell_help ();
