@@ -26,6 +26,7 @@
 #include <nettle/rsa.h>
 #include <nettle/base64.h>
 #include <nettle/asn1.h>
+#include <nettle/bignum.h>
 #include "mailfromd.h"
 #include <mailutils/imaputil.h>
 #include "dkim.h"
@@ -48,7 +49,7 @@ enum {
 	READ_PEM_EOF
 };
 
-/* Returns READ_PEM_OK on match. */ 
+/* Returns READ_PEM_OK on match. */
 static int
 match_pem_start(size_t length, const uint8_t *line,
 		size_t *marker_start,
@@ -70,7 +71,7 @@ match_pem_start(size_t length, const uint8_t *line,
 }
 
 /* Returns READ_PEM_OK on match, READ_PEM_EOF if the line is of the right
-   form except for the marker, otherwise READ_PEM_ERROR. */ 
+   form except for the marker, otherwise READ_PEM_ERROR. */
 static int
 match_pem_end(size_t length, const uint8_t *line,
 	      size_t marker_length,
@@ -92,7 +93,7 @@ match_pem_end(size_t length, const uint8_t *line,
 			return READ_PEM_EOF;
 	}
 
-	return READ_PEM_ERROR;  
+	return READ_PEM_ERROR;
 }
 
 struct pem_info {
@@ -107,7 +108,7 @@ static int
 read_line(FILE *fp, struct nettle_buffer *buffer)
 {
 	int c;
-  
+
 	while ((c = getc(fp)) != EOF) {
 		if (!NETTLE_BUFFER_PUTC(buffer, c))
 			return READ_PEM_ERROR;
@@ -126,10 +127,10 @@ read_line(FILE *fp, struct nettle_buffer *buffer)
  * fp          input file.
  * buffer      buffer to read PEM into.
  * info        fill this structure with information about PEM structure.
- */             
+ */
 static int
 read_pem(FILE *fp, struct nettle_buffer *buffer, struct pem_info *info)
-{  
+{
 	int rc;
 
 	/* Find start line */
@@ -148,10 +149,10 @@ read_pem(FILE *fp, struct nettle_buffer *buffer, struct pem_info *info)
 	buffer->contents[info->marker_start + info->marker_length] = 0;
 
 	info->data_start = buffer->size;
-	
+
 	for (;;) {
 		size_t line_start = buffer->size;
-		
+
 		if ((rc = read_line(fp, buffer)) != READ_PEM_OK)
 			return rc;
 
@@ -184,9 +185,9 @@ static int
 decode_base64(struct nettle_buffer *buffer, size_t start, size_t *length)
 {
 	struct base64_decode_ctx ctx;
-  
+
 	base64_decode_init(&ctx);
-	
+
 	/* Decode in place */
 	if (base64_decode_in_place(&ctx, length, *length,
 				   buffer->contents + start)
@@ -230,7 +231,7 @@ read_keys(FILE *fp, struct rsa_public_key *pub, struct rsa_private_key *priv)
 	else
 		rc = READ_PEM_ERROR;
 	nettle_buffer_clear(&buffer);
-	
+
 	return rc;
 }
 
@@ -241,22 +242,22 @@ pubkey_from_base64(struct rsa_public_key *pub, const char *str)
 	size_t length = strlen(str);
 	struct asn1_der_iterator i, j;
 	int result = READ_PEM_ERROR;
-	
+
 	nettle_buffer_init_realloc(&buffer, NULL, nettle_xrealloc);
 	nettle_buffer_write(&buffer, strlen(str), (const uint8_t*) str);
 	if (decode_base64(&buffer, 0, &length) == READ_PEM_OK
 
 	/* SubjectPublicKeyInfo ::= SEQUENCE {
 	       algorithm		AlgorithmIdentifier,
-	       subjectPublicKey 	BIT STRING
+	       subjectPublicKey		BIT STRING
 	   }
-	   
+
 	   AlgorithmIdentifier ::= SEQUENCE {
-	       algorithm  	OBJECT IDENTIFIER,
-	       parameters 	OPTIONAL
+	       algorithm	OBJECT IDENTIFIER,
+	       parameters	OPTIONAL
 	   }
 	*/
-	
+
 	    && asn1_der_iterator_first(&i, length, buffer.contents) == ASN1_ITERATOR_CONSTRUCTED
 	    && i.type == ASN1_SEQUENCE
 	    && asn1_der_decode_constructed_last(&i) == ASN1_ITERATOR_CONSTRUCTED
@@ -330,7 +331,7 @@ dkim_header_list_next(void *save)
 {
 	struct h_list_buf *hbuf = save;
 	char const *hp = hbuf->ptr;
-	
+
 	while (*hp && (*hp == ' ' || *hp == '\t' || *hp == ':'))
 		hp++;
 	if (*hp) {
@@ -375,12 +376,12 @@ dkim_header_list_match(char const *h_list, char const *h)
 	size_t len = strlen (h);
 	while (*h_list) {
 		size_t n;
-		
+
 		while (*h_list && (*h_list == ' ' || *h_list == '\t'))
 			h_list++;
 		if (*h_list == 0)
 			break;
-		
+
 		n = strcspn(h_list, " \t:");
 		if (n == len && mu_c_strncasecmp (h_list, h, len) == 0)
 			return 1;
@@ -395,6 +396,128 @@ dkim_header_list_match(char const *h_list, char const *h)
 	return 0;
 }
 
+char const *dkim_algo_str[] = { "rsa-sha1", "rsa-sha256" };
+
+#define DKIM_DIGEST_BUFFER_SIZE \
+  (sizeof(union { char s1[SHA1_DIGEST_SIZE]; char s2[SHA256_DIGEST_SIZE]; }))
+
+struct dkim_ctx {
+	int type;
+	union {
+		struct sha256_ctx sha256;
+		struct sha1_ctx sha1;
+	} ctx;
+};
+
+int
+dkim_algo(char const *a)
+{
+	int i;
+
+	for (i = 0; i < sizeof(dkim_algo_str)/sizeof(dkim_algo_str[0]); i++)
+		if (strcmp(a, dkim_algo_str[i]) == 0)
+			return i;
+	return -1;
+}
+
+static void
+dkim_ctx_init(struct dkim_ctx *ctx, int type)
+{
+	ctx->type = type;
+	switch (ctx->type) {
+	case DKIM_ALGO_RSA_SHA1:
+		sha1_init(&ctx->ctx.sha1);
+		break;
+
+	case DKIM_ALGO_RSA_SHA256:
+		sha256_init(&ctx->ctx.sha256);
+		break;
+
+	default:
+		abort();
+	}
+}
+
+static size_t
+dkim_ctx_digest(struct dkim_ctx *ctx, uint8_t *data)
+{
+	size_t size;
+
+	switch (ctx->type) {
+	case DKIM_ALGO_RSA_SHA1:
+		size = SHA1_DIGEST_SIZE;
+		sha1_digest(&ctx->ctx.sha1, size, data);
+		break;
+
+	case DKIM_ALGO_RSA_SHA256:
+		size = SHA256_DIGEST_SIZE;
+		sha256_digest(&ctx->ctx.sha256, size, data);
+		break;
+
+	default:
+		abort();
+	}
+	return size;
+}
+
+static void
+dkim_ctx_update(struct dkim_ctx *ctx, size_t length, const uint8_t *data)
+{
+	switch (ctx->type) {
+	case DKIM_ALGO_RSA_SHA1:
+		sha1_update(&ctx->ctx.sha1, length, data);
+		break;
+
+	case DKIM_ALGO_RSA_SHA256:
+		sha256_update(&ctx->ctx.sha256, length, data);
+		break;
+
+	default:
+		abort();
+	}
+}
+
+static int
+dkim_ctx_sign(const struct rsa_private_key *key, struct dkim_ctx *hash, mpz_t signature)
+{
+	int rc;
+
+	switch (hash->type) {
+	case DKIM_ALGO_RSA_SHA1:
+		rc = rsa_sha1_sign(key, &hash->ctx.sha1, signature);
+		break;
+
+	case DKIM_ALGO_RSA_SHA256:
+		rc = rsa_sha256_sign(key, &hash->ctx.sha256, signature);
+		break;
+
+	default:
+		abort();
+	}
+	return rc;
+}
+
+static int
+dkim_ctx_verify(const struct rsa_public_key *key,
+		struct dkim_ctx *hash, const mpz_t signature)
+{
+	int rc;
+
+	switch (hash->type) {
+	case DKIM_ALGO_RSA_SHA1:
+		rc = rsa_sha1_verify(key, &hash->ctx.sha1, signature);
+		break;
+
+	case DKIM_ALGO_RSA_SHA256:
+		rc = rsa_sha256_verify(key, &hash->ctx.sha256, signature);
+		break;
+
+	default:
+		abort();
+	}
+	return rc;
+}
+
 /*
  * SHA256 hashing functions.
  */
@@ -405,29 +528,29 @@ dkim_header_list_match(char const *h_list, char const *h)
  * the current position and up to the end of file.
  */
 static int
-hash_stream(mu_stream_t str, struct sha256_ctx *ctx)
+hash_stream(mu_stream_t str, struct dkim_ctx *ctx)
 {
 	uint8_t buffer[BUF_SIZE];
 	size_t count;
 	int rc;
-	
+
 	while ((rc = mu_stream_read(str, buffer, sizeof(buffer), &count)) == 0
 	       && count > 0) {
 //		mu_error("READ %*.*s", (int)count, (int)count, buffer);
-		sha256_update(ctx, count, buffer);
+		dkim_ctx_update(ctx, count, buffer);
 	}
 
 	if (rc) {
 		mu_diag_funcall(MU_DIAG_ERROR, "mu_stream_read", NULL, rc);
 		return -1;
 	}
-	
+
 	return 0;
 }
 
 /* Hash LEN bytes from the current position in stream STR. */
 static int
-hash_stream_segment(mu_stream_t str, size_t len, struct sha256_ctx *ctx)
+hash_stream_segment(mu_stream_t str, size_t len, struct dkim_ctx *ctx)
 {
 	while (len) {
 		uint8_t buffer[BUF_SIZE];
@@ -446,7 +569,7 @@ hash_stream_segment(mu_stream_t str, size_t len, struct sha256_ctx *ctx)
 			mu_error(_("unexpected end of file in canonical stream"));
 			return -1;
 		}
-		sha256_update(ctx, count, buffer);
+		dkim_ctx_update(ctx, count, buffer);
 		len -= count;
 	}
 
@@ -455,23 +578,24 @@ hash_stream_segment(mu_stream_t str, size_t len, struct sha256_ctx *ctx)
 
 /* Hash the remaining content of the stream from the current position
  * and store a nul-terminated base64 encoded result in outbuf.
- * Outbuf must have at least BASE64_ENCODE_RAW_LENGTH(SHA256_DIGEST_SIZE) + 1
+ * Outbuf must have at least BASE64_ENCODE_RAW_LENGTH(DKIM_DIGEST_BUFFER_SIZE)
  * bytes of capacity,
  */
 static int
-dkim_body_hash(mu_stream_t str, size_t len, uint8_t *outbuf)
+dkim_body_hash(mu_stream_t str, size_t len, int type, uint8_t *outbuf, size_t *outlen)
 {
-	struct sha256_ctx ctx;
-	uint8_t body_digest[SHA256_DIGEST_SIZE];
-	
-	sha256_init(&ctx);
+	struct dkim_ctx ctx;
+	uint8_t body_digest[DKIM_DIGEST_BUFFER_SIZE];
+	size_t ds;
+
+	dkim_ctx_init(&ctx, type);
 	if (len == DKIM_LENGTH_ALL)
 		hash_stream(str, &ctx);
 	else
 		hash_stream_segment(str, len, &ctx);
-	sha256_digest(&ctx, sizeof(body_digest), body_digest);
-	base64_encode_raw(outbuf, sizeof(body_digest), body_digest);
-
+	ds = dkim_ctx_digest(&ctx, body_digest);
+	base64_encode_raw(outbuf, ds, body_digest);
+	*outlen = BASE64_ENCODE_RAW_LENGTH(ds);
 	return 0;
 }
 
@@ -483,17 +607,17 @@ dkim_body_hash(mu_stream_t str, size_t len, uint8_t *outbuf)
  * RSA digest in base64.  Return value is stored in RET_B64 (malloced).
  */
 int
-dkim_rsa_sha256_sign(struct rsa_private_key *priv, struct sha256_ctx *ctx,
-		     uint8_t **ret_b64)
+dkim_digest(struct rsa_private_key *priv, struct dkim_ctx *ctx,
+	    uint8_t **ret_b64)
 {
 	int rc;
 	mpz_t sig;
 	size_t i;
 	struct nettle_buffer buffer;
 	uint8_t *outbuf;
-	
+
 	mpz_init(sig);
-	rc = rsa_sha256_sign(priv, ctx, sig);
+	rc = dkim_ctx_sign(priv, ctx, sig);
 	if (!rc)
 		return -1;
 
@@ -515,11 +639,11 @@ dkim_rsa_sha256_sign(struct rsa_private_key *priv, struct sha256_ctx *ctx,
 		nettle_buffer_clear(&buffer);
 		return -1;
 	}
-	
+
 	base64_encode_raw(outbuf, buffer.size, buffer.contents);
 	outbuf[BASE64_ENCODE_RAW_LENGTH(buffer.size)] = 0;
 	nettle_buffer_clear(&buffer);
-	
+
 	*ret_b64 = outbuf;
 
 	return 0;
@@ -530,7 +654,13 @@ dkim_rsa_sha256_sign(struct rsa_private_key *priv, struct sha256_ctx *ctx,
  * re. "DKIM-Signature Canonicalization Header" IANA registry.
  */
 static char const *dkim_canon_string[] = { "simple", "relaxed", NULL };
-#define DKIM_CANON_STRING_MAX (sizeof(dkim_canon_string[DKIM_CANON_RELAXED])-1)
+
+/*
+ * Buffer length for holding the longest possible canonicalization
+ * header.
+ */
+#define DKIM_CANON_STRING_MAX sizeof("relaxed/relaxed")
+
 /*
  * Convert STR to a DKIM_CANON_* constant (return DKIM_CANON_ERR on error).
  * If ENDP is NULL, STR must be one of the strings from dkim_canon_string.
@@ -604,12 +734,12 @@ dkim_format_tag_base64(struct dkim_format_buf *fb, char const *tag,
 		       char const *val)
 {
 	size_t len = strlen(val);
-	
+
 	if (!dkim_format_wrap(fb, strlen(tag) + strlen(val) + 2))
 		dkim_format_wrap(fb, strlen(tag) + 1);
 	mu_stream_write(fb->str, tag, strlen(tag), NULL);
 	mu_stream_write(fb->str, "=", 1, NULL);
-	
+
 	while (len > 0) {
 		size_t n = MAX_LINE_LEN - fb->stat[MU_STREAM_STAT_OUT];
 		if (len < n)
@@ -681,7 +811,7 @@ dkim_tag_a_formatter(struct dkim_format_buf *fb,
 		     void *data)
 {
 	char **cptr = data;
-	dkim_format_tag(fb, tag, *cptr ? *cptr : DKIM_ALGORITHM);
+	dkim_format_tag(fb, tag, *cptr ? *cptr : dkim_algo_str[DKIM_ALGO_DEFAULT]);
 }
 
 /* Formatter for the q= tag */
@@ -703,10 +833,10 @@ dkim_tag_h_formatter(struct dkim_format_buf *fb,
 	char **cptr = data;
 	void *save;
 	char const *hval;
-	
+
 	if (!dkim_format_wrap(fb, 2))
 		dkim_format_wrap(fb, strlen(*cptr) + 3);
-		
+
 	mu_stream_write(fb->str, "h=", 2, NULL);
 
 	if ((hval = dkim_header_list_first(*cptr, &save)) != NULL) {
@@ -729,7 +859,7 @@ dkim_tag_c_parser(char const *value, void *data)
 {
 	int *canon = data;
 	char *s;
-			
+
 	if ((canon[0] = dkim_str_to_canon_type(value, &s)) == DKIM_CANON_ERR)
 		return -1;
 	if (*s == 0)
@@ -746,7 +876,7 @@ dkim_tag_c_formatter(struct dkim_format_buf *fb,
 		     void *data)
 {
 	int *canon = data;
-	char v[(DKIM_CANON_STRING_MAX+1)*2];
+	char v[DKIM_CANON_STRING_MAX];
 	strcpy(v, dkim_canon_string[canon[0]]);
 	strcat(v, "/");
 	strcat(v, dkim_canon_string[canon[1]]);
@@ -791,7 +921,7 @@ dkim_tag_l_parser(char const *value, void *data)
 	char *endp;
 	errno = 0;
 	n = strtoul(value, &endp, 10);
-	if (errno || *endp) 
+	if (errno || *endp)
 		return -1;
 	*sptr = n;
 	return 0;
@@ -836,9 +966,9 @@ struct dkim_tag_descr {
 	char *tag;       /* Tag name. */
 	size_t off;      /* Field offset in struct dkim_signature. */
 	int (*parser)(char const *, void *);
-       	                 /* Parser function. */
+			 /* Parser function. */
 	void (*formatter)(struct dkim_format_buf *, char const *, void *);
-	                 /* Formatter function. */
+			 /* Formatter function. */
 };
 
 /* Order of entries in this array defines the order in which tags are
@@ -881,7 +1011,7 @@ static struct dkim_tag_descr tag_descr[] = {
 		dkim_tag_char_parser,
 		dkim_tag_q_formatter
 	},
-	{		
+	{
 		"h",
 		offsetof(struct dkim_signature, h),
 		dkim_tag_char_parser,
@@ -892,7 +1022,7 @@ static struct dkim_tag_descr tag_descr[] = {
 		offsetof(struct dkim_signature, i),
 		dkim_tag_char_parser,
 		dkim_tag_char_formatter,
-	},		
+	},
 	{
 		"l",
 		offsetof(struct dkim_signature, l),
@@ -905,7 +1035,7 @@ static struct dkim_tag_descr tag_descr[] = {
 		dkim_tag_time_parser,
 		dkim_tag_time_formatter
 	},
-	{		
+	{
 		"x",
 		offsetof(struct dkim_signature, x),
 		dkim_tag_time_parser,
@@ -916,7 +1046,7 @@ static struct dkim_tag_descr tag_descr[] = {
 		offsetof(struct dkim_signature, bh),
 		dkim_tag_char_parser,
 		dkim_tag_bh_formatter
-	},		
+	},
 	{
 		"b",
 		offsetof(struct dkim_signature, b),
@@ -1032,15 +1162,15 @@ dkim_signature_format(struct dkim_signature *sig, int crlf, char **result)
 		//FIXME: assert(size < (size_t)~0)
 		size = off;
 		text = mu_alloc(size + 1);
-		
+
 		mu_stream_seek(fb.str, 0, MU_SEEK_SET, NULL);
-		
+
 		mu_stream_read(fb.str, text, size, NULL);
 		text[size] = 0;
 		*result = text;
 	}
 	mu_stream_destroy(&fb.str);
-	
+
 	return rc;
 }
 
@@ -1056,7 +1186,7 @@ canonicalize(mu_message_t msg, int canon[2], mu_stream_t *canon_str)
 {
 	mu_stream_t mstr, flt, in;
 	int rc;
-	
+
 	rc = mu_temp_file_stream_create(&mstr, NULL, 0);
 	if (rc) {
 		mu_diag_funcall(MU_DIAG_ERROR,
@@ -1073,7 +1203,7 @@ canonicalize(mu_message_t msg, int canon[2], mu_stream_t *canon_str)
 		mu_stream_destroy(&mstr);
 		return -1;
 	}
-	
+
 	rc = dkim_canonicalizer_create(&flt, in,
 				       canon[0],
 				       canon[1],
@@ -1085,7 +1215,7 @@ canonicalize(mu_message_t msg, int canon[2], mu_stream_t *canon_str)
 		mu_stream_destroy(&mstr);
 		return -1;
 	}
-		
+
 	in = flt;
 	rc = mu_filter_create(&flt, in, "CRLF",
 			      MU_FILTER_ENCODE, MU_STREAM_READ);
@@ -1210,7 +1340,7 @@ dkim_tag_find(char const *sigstr, char const *tag, size_t *ret_len)
 	int i;
 	size_t tag_len = strlen(tag);
 	size_t sig_len = strlen(sigstr);
-	
+
 	for (i = 0; i + tag_len + 1 < sig_len; i++) {
 		if (!(sigstr[i] == ' ' || sigstr[i] == '\t' ||
 		      sigstr[i] == '\r' || sigstr[i] == '\n')) {
@@ -1232,8 +1362,6 @@ enum {
 	DKIM_HASH_DIFF
 };
 
-#define BH_SIZE (BASE64_ENCODE_RAW_LENGTH(SHA256_DIGEST_SIZE))
-
 /*
  * dkim_hash(MSG, SIG, SIGSTR, CTX)
  * --------------------------------
@@ -1260,7 +1388,7 @@ enum {
  */
 static int
 dkim_hash(mu_message_t msg, struct dkim_signature *sig, char const *sigstr,
-	  struct sha256_ctx *ctx)
+	  struct dkim_ctx *ctx)
 {
 	mu_stream_t canon_stream = NULL;
 	char *hp;
@@ -1268,7 +1396,8 @@ dkim_hash(mu_message_t msg, struct dkim_signature *sig, char const *sigstr,
 	struct header_map h_all = HEADER_MAP_INITIALIZER(h_all);
 	struct header_map h_sel = HEADER_MAP_INITIALIZER(h_sel);
 	struct header_map *hmap;
-	uint8_t bh[BH_SIZE];
+	uint8_t bh[BASE64_ENCODE_RAW_LENGTH(DKIM_DIGEST_BUFFER_SIZE)];
+	size_t bhs;
 	mu_opool_t op = NULL;
 	char c;
 	int rc;
@@ -1277,7 +1406,7 @@ dkim_hash(mu_message_t msg, struct dkim_signature *sig, char const *sigstr,
 	enum { H_INIT, H_HEADER, H_CR1, H_CR2, H_NL } state = H_INIT;
 	mu_stream_t sigcanon, str;
 	char *sig_str_buf;
-	
+
 	/* Create a canonical representation of the message */
 	if (canonicalize(msg, sig->canon, &canon_stream))
 		goto err;
@@ -1346,13 +1475,13 @@ dkim_hash(mu_message_t msg, struct dkim_signature *sig, char const *sigstr,
 	}
 
 	/* Error exit from the above loop */
-	if (rc) 
+	if (rc)
 		mu_diag_funcall(MU_DIAG_ERROR, "mu_stream_read", NULL, rc);
 badstream:
 	mu_error(_("malformed canonical stream"));
 	goto err;
 
-end:	
+end:
 	/*
 	 * Scanning terminated successfully.  Current position in the
 	 * canon_stream is left at the start of the body, which will come
@@ -1384,21 +1513,21 @@ end:
 	dkim_header_list_end(hstate);
 
 	/* Hash the body */
-	if (dkim_body_hash(canon_stream, sig->l, bh))
+	if (dkim_body_hash(canon_stream, sig->l, sig->algo, bh, &bhs))
 		goto err;
 	if (sig->bh) {
-		if (memcmp(sig->bh, bh, BH_SIZE)) {
+		if (memcmp(sig->bh, bh, bhs)) {
 			result = DKIM_HASH_DIFF;
 			goto err;
 		}
 	} else {
-		sig->bh = malloc(BH_SIZE + 1);
+		sig->bh = malloc(bhs + 1);
 		if (!sig->bh)
 			goto err;
-		memcpy(sig->bh, bh, BH_SIZE);
-		sig->bh[BH_SIZE] = 0;
+		memcpy(sig->bh, bh, bhs);
+		sig->bh[bhs] = 0;
 	}
-	
+
 	/* Hash the selected headers */
 	HEADER_MAP_FOREACH(hmap, &h_sel) {
 		mu_stream_seek(canon_stream, hmap->start, MU_SEEK_SET, NULL);
@@ -1446,7 +1575,7 @@ end:
 		goto err;
 	}
 	sigcanon = str;
-	
+
 	hash_stream(sigcanon, ctx);
 	mu_stream_unref(sigcanon);
 	free(sig_str_buf);
@@ -1459,7 +1588,7 @@ err:
 	mu_stream_destroy(&canon_stream);
 	return result;
 }
-	
+
 
 /* Sign the message.  Arguments:
  *
@@ -1482,9 +1611,9 @@ mfd_dkim_sign(mu_message_t msg, struct dkim_signature *sig,
 	struct rsa_private_key priv;
 	struct rsa_public_key pub;
 	FILE *fp;
-	struct sha256_ctx ctx;
+	struct dkim_ctx ctx;
 	int result = -1;
-	
+
 	fp = fopen(priv_file, "r");
 	if (!fp) {
 		mu_error(_("can't open %s: %s"), priv_file, strerror(errno));
@@ -1500,14 +1629,14 @@ mfd_dkim_sign(mu_message_t msg, struct dkim_signature *sig,
 	}
 	rsa_public_key_clear(&pub);
 
-	sha256_init(&ctx);
+	dkim_ctx_init(&ctx, sig->algo);
 	if (dkim_hash(msg, sig, NULL, &ctx) == DKIM_HASH_OK) {
-	    /* Create the RSA-SHA256 signature in b */
-	    dkim_rsa_sha256_sign(&priv, &ctx, &sig->b);
-	
-	    /* Create the header */
-	    dkim_signature_format(sig, 1, ret_sighdr);
-	    result = 0;
+		/* Create the RSA-SHA256 signature in b */
+		dkim_digest(&priv, &ctx, &sig->b);
+
+		/* Create the header */
+		dkim_signature_format(sig, 1, ret_sighdr);
+		result = 0;
 	}
 	/* Reclaim the allocated memory. */
 	free(sig->b);
@@ -1526,7 +1655,7 @@ char const *dkim_explanation_str[] = {
 	[DKIM_EXPL_SIG_MISS] = "signature is missing required tag",
 	[DKIM_EXPL_DOMAIN_MISMATCH] = "domain mismatch",
 	[DKIM_EXPL_BAD_VERSION] = "incompatible version",
-	[DKIM_EXPL_BAD_ALGORITHM] = "unsupported algorithm",
+	[DKIM_EXPL_BAD_ALGORITHM] = "unsupported signing algorithm",
 	[DKIM_EXPL_BAD_QUERY] = "unsupported query method",
 	[DKIM_EXPL_FROM] = "From field not signed",
 	[DKIM_EXPL_EXPIRED] = "signature expired",
@@ -1537,6 +1666,7 @@ char const *dkim_explanation_str[] = {
 	[DKIM_EXPL_BAD_BODY] = "body hash did not verify",
 	[DKIM_EXPL_BAD_BASE64] = "can't decode b= tag",
 	[DKIM_EXPL_BAD_SIG] = "signature did not verify",
+	[DKIM_EXPL_BAD_KEY_TYPE] = "unsupported public key type",
 };
 
 int dkim_result_trans[] = {
@@ -1563,6 +1693,8 @@ int dkim_result_trans[] = {
 static int
 dkim_sig_validate(struct dkim_signature *sig)
 {
+	int algo;
+
 	if (!sig->a
 	    || !sig->b
 	    || !sig->bh
@@ -1572,18 +1704,19 @@ dkim_sig_validate(struct dkim_signature *sig)
 	    || !sig->v) {
 		return DKIM_EXPL_SIG_MISS;
 	}
-	    
+
 	if (strcmp(sig->v, DKIM_VERSION))
 		return DKIM_EXPL_BAD_VERSION;
 
-	if (strcmp(sig->a, DKIM_ALGORITHM))
+	if ((algo = dkim_algo(sig->a)) == -1)
 		return DKIM_EXPL_BAD_ALGORITHM;
-	
-        if (!sig->q)
+	sig->algo = algo;
+
+	if (!sig->q)
 		sig->q = mu_strdup(DKIM_QUERY_METHOD);
 	else if (strcmp(sig->q, DKIM_QUERY_METHOD))
 		return DKIM_EXPL_BAD_QUERY;
-	
+
 	if (sig->i) {
 		char *p = strchr(sig->i, '@');
 		size_t ilen, dlen;
@@ -1597,7 +1730,7 @@ dkim_sig_validate(struct dkim_signature *sig)
 		      (p[ilen - dlen - 1] == '.' || p[ilen - dlen - 1] == '@')))
 			return DKIM_EXPL_DOMAIN_MISMATCH;
 	}
-	
+
 	if (!dkim_header_list_match(sig->h, MU_HEADER_FROM))
 		return DKIM_EXPL_FROM;
 
@@ -1613,11 +1746,11 @@ dnsrec_parse(char *rec, mu_assoc_t *pa)
 	mu_assoc_t a;
 	struct mu_wordsplit ws;
 	int result;
-	
+
 	if (mu_assoc_create (&a, 0))
 		mu_alloc_die ();
 	mu_assoc_set_destroy_item (a, mu_list_free_item);
-		
+
 	ws.ws_delim = ";";
 	if (mu_wordsplit(rec,
 			 &ws,
@@ -1628,12 +1761,12 @@ dnsrec_parse(char *rec, mu_assoc_t *pa)
 		result = 1;
 	} else {
 		size_t i;
-		
+
 		for (i = 0; i < ws.ws_wordc; i++) {
 			char *p = strchr(ws.ws_wordv[i], '=');
 			char **slot;
 			int rc;
-				
+
 			if (!p) {
 				result = 1;
 				break;
@@ -1641,7 +1774,7 @@ dnsrec_parse(char *rec, mu_assoc_t *pa)
 			*p++ = 0;
 			mu_rtrim_class(ws.ws_wordv[i], MU_CTYPE_BLANK);
 			mu_ltrim_class(p, MU_CTYPE_BLANK);
-			
+
 			rc = mu_assoc_install_ref(a, ws.ws_wordv[i], &slot);
 			if (rc == ENOMEM)
 				mu_alloc_die ();
@@ -1653,7 +1786,7 @@ dnsrec_parse(char *rec, mu_assoc_t *pa)
 			*slot = mu_strdup(p);
 		}
 	}
-		
+
 	mu_wordsplit_free(&ws);
 	if (result)
 		mu_assoc_destroy (&a);
@@ -1667,22 +1800,21 @@ pubkey_validate(mu_assoc_t a, struct dkim_signature const *sig)
 {
 	char *s;
 	size_t n;
-	
+
 	if ((s = mu_assoc_get(a, "v")) != NULL &&
 	    strcmp(s, DKIM_KEYRECORD_VERSION))
 		return DKIM_EXPL_KEY_SYNTAX;
 
-	if ((s = mu_assoc_get(a, "p")) == NULL) 
+	if ((s = mu_assoc_get(a, "p")) == NULL)
 		return DKIM_EXPL_KEY_SYNTAX;
-	
+
 	if (s[0] == 0)
 		return DKIM_EXPL_KEY_REVOKED;
 
-	n = strcspn(sig->a, "-");
-	if ((s = mu_assoc_get(a, "k")) != NULL &&
-	    !(n == strlen(s) && memcmp(s, sig->a, n) == 0))
-		return DKIM_EXPL_BAD_ALGORITHM;
+	if ((s = mu_assoc_get(a, "k")) != NULL && strcmp(s, DKIM_KEY_TYPE))
+		return DKIM_EXPL_BAD_KEY_TYPE;
 
+	n = strcspn(sig->a, "-");
 	if ((s = mu_assoc_get(a, "h")) != NULL &&
 	    !dkim_header_list_match(s, sig->a + n + 1))
 		return DKIM_EXPL_BAD_ALGORITHM;
@@ -1691,9 +1823,10 @@ pubkey_validate(mu_assoc_t a, struct dkim_signature const *sig)
 
 static int
 dkim_sig_key_verify(mu_message_t msg, struct dkim_signature *sig,
+		    int type,
 		    struct rsa_public_key *pub)
 {
-	struct sha256_ctx ctx;
+	struct dkim_ctx ctx;
 	mpz_t bs;
 	int rc;
 	struct nettle_buffer buffer;
@@ -1714,8 +1847,8 @@ dkim_sig_key_verify(mu_message_t msg, struct dkim_signature *sig,
 				DKIM_SIGNATURE_HEADER, rc);
 		return DKIM_EXPL_INTERNAL_ERROR;
 	}
-		
-	sha256_init(&ctx);
+
+	dkim_ctx_init(&ctx, type);
 	switch (dkim_hash(msg, sig, sigstr, &ctx)) {
 	case DKIM_HASH_OK:
 		break;
@@ -1725,7 +1858,7 @@ dkim_sig_key_verify(mu_message_t msg, struct dkim_signature *sig,
 
 	case DKIM_HASH_ERR:
 		return DKIM_EXPL_INTERNAL_ERROR;
-	}		
+	}
 
 	length = strlen((char*)sig->b);
 	nettle_buffer_init_realloc(&buffer, NULL, nettle_xrealloc);
@@ -1734,10 +1867,10 @@ dkim_sig_key_verify(mu_message_t msg, struct dkim_signature *sig,
 		return DKIM_EXPL_BAD_BASE64;
 
 	nettle_mpz_init_set_str_256_u(bs, length, buffer.contents);
-	rc = rsa_sha256_verify (pub, &ctx, bs);
+	rc = dkim_ctx_verify(pub, &ctx, bs);
 	nettle_buffer_clear(&buffer);
 	mpz_clear(bs);
-	
+
 	return rc ? DKIM_EXPL_OK : DKIM_EXPL_BAD_SIG;
 }
 
@@ -1747,7 +1880,7 @@ dkim_sig_verify(mu_message_t msg, struct dkim_signature *sig)
 	char **dnsrec;
 	int i;
 	int result;
-	
+
 	/* Get the DKIM DNS record */
 	switch (dkim_lookup(sig->d, sig->s, &dnsrec)) {
 	case dns_success:
@@ -1764,7 +1897,7 @@ dkim_sig_verify(mu_message_t msg, struct dkim_signature *sig)
 		mu_assoc_t a;
 		struct rsa_public_key pub;
 		int rc;
-		
+
 		if (dnsrec_parse(dnsrec[i], &a))
 			continue;
 
@@ -1774,7 +1907,7 @@ dkim_sig_verify(mu_message_t msg, struct dkim_signature *sig)
 			   != READ_PEM_OK) {
 			result = DKIM_EXPL_KEY_SYNTAX;
 		} else {
-			result = dkim_sig_key_verify(msg, sig, &pub);
+			result = dkim_sig_key_verify(msg, sig, sig->algo, &pub);
 			rsa_public_key_clear(&pub);
 		}
 		mu_assoc_destroy(&a);
@@ -1786,7 +1919,7 @@ dkim_sig_verify(mu_message_t msg, struct dkim_signature *sig)
 	for (i = 0; dnsrec[i]; i++)
 		free(dnsrec[i]);
 	free(dnsrec);
-	
+
 	return result;
 }
 
@@ -1807,13 +1940,17 @@ wselim(char *s)
 }
 
 int
-mfd_dkim_verify(mu_message_t msg, char **ret_sig)
+mfd_dkim_verify(mu_message_t msg, char **ret_sig, char const **algo)
 {
 	mu_header_t hdr;
 	int rc;
 	int result = DKIM_EXPL_NO_SIG;
 	size_t i;
-	
+
+	if (algo) {
+		*algo = NULL;
+	}
+
 	/* Get the DKIM-Signature header from the message */
 	rc = mu_message_get_header(msg, &hdr);
 	if (rc) {
@@ -1825,7 +1962,7 @@ mfd_dkim_verify(mu_message_t msg, char **ret_sig)
 	for (i = 1; result != DKIM_EXPL_OK; i++) {
 		struct dkim_signature sig;
 		char *sig_str;
-		
+
 		rc = mu_header_aget_value_unfold_n(hdr,
 						   DKIM_SIGNATURE_HEADER,
 						   i,
@@ -1847,11 +1984,17 @@ mfd_dkim_verify(mu_message_t msg, char **ret_sig)
 		} else {
 			/* Validate the signature */
 			result = dkim_sig_validate(&sig);
-			if (result == DKIM_EXPL_OK)
+			if (result == DKIM_EXPL_OK) {
+				if (algo) {
+					*algo = dkim_algo_str[sig.algo];
+				}
 				result = dkim_sig_verify(msg, &sig);
-			if (result == DKIM_EXPL_OK && ret_sig) {
-				*ret_sig = sig_str;
-				sig_str = NULL;
+			}
+			if (result == DKIM_EXPL_OK) {
+				if (ret_sig) {
+					*ret_sig = sig_str;
+					sig_str = NULL;
+				}
 			}
 			dkim_signature_free(&sig);
 		}
@@ -1859,4 +2002,3 @@ mfd_dkim_verify(mu_message_t msg, char **ret_sig)
 	}
 	return result;
 }
-
