@@ -265,6 +265,7 @@ struct environ_locus {
 };
 
 struct environ_cleanup_closure {
+	int when;
 	void (*cleanup)(void *);
 	void *data;
 };
@@ -286,6 +287,7 @@ struct eval_environ {
 	struct environ_locus locus;
 	
 	/* Temporary heap space */
+	int temp_active;
 	size_t temp_start;
 	size_t temp_size;
 
@@ -328,8 +330,6 @@ struct eval_environ {
 	/* Function clean-up sequence */
 	mu_list_t function_cleanup_list;
 
-	mu_list_t environ_cleanup_list;
-	
 	/* Exit information */
 	sfsistat status;       /* Program exit status */
 	jmp_buf x_jmp;         /* Return point for runtime errors */
@@ -448,15 +448,7 @@ _cleanup_compare(const void *a, const void *b)
 {
 	struct environ_cleanup_closure const *closa = a;
 	struct environ_cleanup_closure const *closb = b;
-	return closa->data != closb->data;
-}
-	
-static void
-_cleanup_destroy(void *item)
-{
-	struct environ_cleanup_closure *clos = item;
-	if (clos->cleanup)
-		clos->cleanup(clos->data);
+	return !(closa->when == closb->when && closa->data == closb->data);
 }
 
 static void
@@ -464,39 +456,49 @@ env_cleanup_list_create(mu_list_t *plist)
 {
 	mu_list_t list;
 	mu_list_create(&list);//FIXME: error checking
-	mu_list_set_destroy_item(list, _cleanup_destroy);
 	mu_list_set_comparator(list, _cleanup_compare);
 	*plist = list;
 }
 
 static void
-env_cleanup_list_append(mu_list_t list, void *data, void (*func)(void *))
+env_cleanup_list_append(mu_list_t list, int when, void *data, void (*func)(void *))
 {
 	struct environ_cleanup_closure *clos = mu_zalloc(sizeof(*clos));
 
+	clos->when = when;
 	clos->cleanup = func ? func : free;
 	clos->data = data;
 	mu_list_append(list, clos);
 }
 
-void
-env_function_cleanup_flush(eval_environ_t env, void *ptr)
+static int
+_cleanup_flush (void *item, void *data)
 {
-	if (ptr) {
-		struct environ_cleanup_closure clos;
-		clos.data = ptr;
-		mu_list_remove(env->function_cleanup_list, &clos);
-	} else {
-		mu_list_clear(env->function_cleanup_list);
+	struct environ_cleanup_closure *clos = item;
+	int when = *(int*)data;
+
+	if (when == CLEANUP_ALWAYS ||
+	    clos->when == CLEANUP_ALWAYS ||
+	    clos->when == when) {
+		clos->cleanup(clos->data);
 	}
+	return 0;
+}
+
+void
+env_function_cleanup_flush(eval_environ_t env, int when)
+{
+	mu_list_foreach(env->function_cleanup_list, _cleanup_flush, &when);
+	mu_list_clear(env->function_cleanup_list);
 }			
 
 void
-env_function_cleanup_del(eval_environ_t env, void *ptr)
+env_function_cleanup_del(eval_environ_t env, int when, void *ptr)
 {
 	int rc;
 	struct environ_cleanup_closure clos, *cptr;
 
+	clos.when = when;
 	clos.data = ptr;
 
 	rc = mu_list_locate(env->function_cleanup_list, &clos, (void**)&cptr);
@@ -507,9 +509,9 @@ env_function_cleanup_del(eval_environ_t env, void *ptr)
 }
 
 void
-env_function_cleanup_add(eval_environ_t env, void *data, void (*func)(void *))
+env_function_cleanup_add(eval_environ_t env, int when, void *data, void (*func)(void *))
 {
-	env_cleanup_list_append(env->function_cleanup_list, data, func);
+	env_cleanup_list_append(env->function_cleanup_list, when, data, func);
 }
 
 /* A call to expand_dataseg (see below) invalidates any C variables that
@@ -820,14 +822,18 @@ heap_obstack_begin(eval_environ_t env)
 {
 	env->temp_start = env->toh;
 	env->temp_size = 0;
+	env->temp_active = 1;
 }
 
 void
 heap_obstack_cancel(eval_environ_t env)
 {
-	env->toh = env->temp_start;
-	env->temp_start = 0;
-	env->temp_size = 0;
+	if (env->temp_active) {
+		env->toh = env->temp_start;
+		env->temp_start = 0;
+		env->temp_size = 0;
+		env->temp_active = 0;
+	}
 }
 
 STKVAL
@@ -837,6 +843,7 @@ heap_obstack_finish(eval_environ_t env)
 	mf_c_val(ret, size) = env->temp_start;
 	env->temp_start = 0;
 	env->temp_size = 0;
+	env->temp_active = 0;
 	return ret;
 }
 
@@ -909,6 +916,26 @@ heap_obstack_grow(eval_environ_t env, void * MFL_DATASEG ptr, size_t size)
 	return ret;
 }
 
+void
+heap_obstack_reclaim(eval_environ_t env, size_t size)
+{
+	if (size > env->temp_size)
+		abort();
+	env->temp_size -= size;
+	env->toh -= B2STACK(size);
+}
+
+void
+heap_obstack_truncate(eval_environ_t env, size_t size)
+{
+	if (size == env->temp_size)
+		return;
+	else if (size > env->temp_size)
+		heap_obstack_grow(env, NULL, size - env->temp_size);
+	else
+		heap_obstack_reclaim(env, env->temp_size - size);
+}
+
 void *
 heap_obstack_base(eval_environ_t env)
 {
@@ -1092,6 +1119,7 @@ instr_stkalloc(eval_environ_t env)
 				      _("Out of stack space; increase #pragma stacksize"));
 	}
 	env->tos -= n;
+	memset(env->dataseg + env->tos, 0, n * sizeof(env->dataseg[0]));
 	advance_pc(env, 1);
 }
 
@@ -2515,7 +2543,7 @@ env_throw_0(eval_environ_t env, mf_exception status, size_t off)
 {
 	prog_counter_t pc;
 
-	env_function_cleanup_flush(env, NULL);
+	env_function_cleanup_flush(env, CLEANUP_THROW);
 	if (status > exception_count)
 		runtime_error(env, _("unknown exception: %d: %s"),
 			      status, (char*) env_data_ref(env, off));
@@ -2557,6 +2585,9 @@ env_throw_bi(eval_environ_t env, mf_exception status, const char *biname,
 {
 	va_list ap;
 	size_t off;
+
+	heap_obstack_cancel(env);
+	
 	va_start(ap, fmt);
 	off = env_vsprintf(env, biname, fmt, ap);
 	va_end(ap);
@@ -2969,7 +3000,6 @@ create_environment(SMFICTX *ctx,
 	env->bi_priv_array = NULL;
 
 	env_cleanup_list_create(&env->function_cleanup_list);
-	env_cleanup_list_create(&env->environ_cleanup_list);
 	
 	if (exception_count) {
 		env->defcatch_ctx = mu_zalloc(exception_count *
@@ -2991,7 +3021,6 @@ destroy_environment(eval_environ_t env)
 	mu_stream_destroy(&env->stream);
 	mu_header_destroy(&env->header);
 	mu_list_destroy(&env->function_cleanup_list);
-	mu_list_destroy(&env->environ_cleanup_list);
 	mu_list_destroy(&env->mmq);
 	env_builtin_priv_destroy(env);
 	free(env);
