@@ -98,6 +98,8 @@ void heap_obstack_begin(eval_environ_t env);
 STKVAL heap_obstack_finish(eval_environ_t env);
 void heap_obstack_cancel(eval_environ_t env);
 void *heap_obstack_grow(eval_environ_t env, void *MFL_DATASEG ptr, size_t size);
+void heap_obstack_reclaim(eval_environ_t env, size_t size);
+void heap_obstack_truncate(eval_environ_t env, size_t size);
 void *heap_obstack_base(eval_environ_t env);
 void heap_obstack_vsprintf(eval_environ_t env, const char *fmt, va_list ap);
 void heap_obstack_sprintf(eval_environ_t env, const char *fmt, ...);
@@ -120,12 +122,19 @@ int builtin_priv_register(void *(*init)(void), void (*destroy)(void*),
 			  void (*free_capture));
 void env_free_captured(eval_environ_t env);
 
-void env_function_cleanup_add(eval_environ_t env, void *data,
+
+enum cleanup_when {
+	CLEANUP_THROW,        /* Cleanup on exception. */
+	CLEANUP_RETURN,       /* Cleanup on return from function. */
+	CLEANUP_ALWAYS        /* Cleanup always. */
+};
+
+void env_function_cleanup_add(eval_environ_t env, int when,
+			      void *data,
 			      void (*func)(void *));
-void env_function_cleanup_del(eval_environ_t env, void *data);
-void env_function_cleanup_flush(eval_environ_t env, void *ptr);
-void env_environ_cleanup_add(eval_environ_t env, void *data,
-			     void (*func)(void *));
+void env_function_cleanup_del(eval_environ_t env, int when, void *data);
+void env_function_cleanup_flush(eval_environ_t env, int when);
+
 
 void ensure_initialized_variable(const char *name, struct value *val);
 void ds_init_variable(const char *name, void *data);
